
extern float quadratic_form(const float qf[3], float phi);
extern void fit_quadratic_form(float qf[3], int N, const float phi[N], const float v[N]);
extern float harmonic(const float qf[3], float phi);
extern void fit_harmonic(float qf[3], int N, const float phi[N], const float v[N]);

