#include "misc/opts.h"

extern struct loss_config_s* get_loss_from_option(void);
extern struct loss_config_s* get_val_loss_from_option(void);

extern struct opt_s loss_opts[];
extern struct opt_s val_loss_opts[];

extern const int N_loss_opts;
extern const int N_val_loss_opts;
