
#ifdef __cplusplus
extern "C" {
#endif

extern void cuda_zrfmac_upper_triagmat(long N, long NC, long NM, long ostr, long istr, long mstr, float* dst, const float* src, const float* mat);

#ifdef __cplusplus
}
#endif
