


extern double bessel_i0(double x);
extern double bessel_i0_compat(double x);
extern double Si_power(double x);
extern double Si(double x);

extern double sinc(double x);
extern float sincf(float x);
extern double jinc(double x);

extern double gamma_func(double x);

extern double hyp2f1(double a, double b, double c, double x);

extern double legendre(double lambda, double x);

extern void roots_weights_gauss_legendre(int N, double mu0, double roots[N], double weights[N]);