
extern void syn(int levels, float sigma[levels], float factors[levels], int nwarps[levels],
	int d, unsigned long flags, int N, const long dims[N],
	_Complex float* disp, _Complex float* idisp,
	const _Complex float* static_img, const _Complex float* moving_img);
	